/*
 * Robert Harrison
 * rharrison (email at sign) hgf.com
 * August 2008
 *
 * Version 0.1 Beta
 *
 * This is a small C program to pass NMEA 0183 output on the GPS serial port 
 * and produce sentences used by the tracker on the UKHAS wiki.
 *
 * NMEA Protocol
 *
 * NMEA data is sent in 8-bit ASCII where the MSB is set to zero (0). 
 * The specification also has a set of reserved characters. These characters 
 * assist in the formatting of the NMEA data string. 
 *
 * The specification also states valid characters and gives a table of 
 * these characters ranging from HEX 20 to HEX 7E.
 *
 * As stated in the NMEA 0183 specification version 3.01 the maximum number 
 * of characters shall be 82, consisting of a maximum of 79 characters between 
 * start of message $ and terminating delimiter <CR><LF> 
 * (HEX 0D and 0A). The minimum number of fields is one (1).
 *
 * Basic sentence format:
 *
 * $aaccc,c--c*hh<CR><LF>
 *
 * $             Start of sentence
 * aaccc         Address field/Command
 * “,”           Field delimiter (Hex 2C)
 * c--c          Data sentence block
 * *             Checksum delimiter (HEX 2A)
 * hh            Checksum field (the hexadecimal value represented in ASCII)
 * <CR><LF>      End of sentence (HEX OD OA)
 *
 *
 * Usage : gpstrack -o -p /dev/???? -f freqency in seconds to report string
 *
 * -i --id		The identification tag to attach to the track string.
 *  			This is used to uniquely identify gps tracking data 
 *  			to an individual unit.
 *  			eg icarus, chase1, chase2 etc...
 *
 * -P --password	The mysql access password to attach to the track 
 *  			string.
 *
 * -p --port    	Set the port to monitor usually /dev/ttySx 
 *  			or /dev/ttyUSBx where x is a number
 *
 * -f --freqency	NOT IMPLEMENTED Frequency of track_string output in
 *                      seconds. Requires gps to be working and providing 
 *                      NMEA strings to work.
 *
 * -e --email		Produce email subject format instead of url
 *  			format.
 *
 * -d --decimal		Use decimal notation for Lat. and Long. rather than
 * 			the default of hhmm.ssss (hours, minutes and seconds)
 *
 * -r --repeat		By default gpstrack generates only one track string
 * 			this option will cause gpstrack to continually 
 * 			generate track strings.
 *
 * TODO
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>

/* 
 * Baudrate settings are defined in <asm/termbits.h>, which is
 * included by <termios.h> 
 */
#define BAUDRATE B4800

#define TRUE 1
#define FALSE 0
#define MAX_NMEA_STRING 82+1

/* Default GPS port */
#define GPSDEVICE "/dev/ttyUSB0"


enum ts_status {ts_ready, ts_pending, ts_printed}; 
enum ts_options {ts_hms_fmt, ts_dec_fmt, ts_url_fmt, ts_email_fmt}; 

struct track_string {
	char id [255];
	char time [20];
	char date [20];
	char lat  [20];
	char lng  [20];
	char alt  [20];
	char hdg  [20];
	char spd  [20];
	char pass [255];
	int fmt; 
	int status;
	};

int ts_pos_format = ts_hms_fmt;
int ts_output_format = ts_url_fmt;
int ts_frequency = 1;
int ts_repeat = FALSE;

char default_nmea_data_port [255] = GPSDEVICE;
char * nmea_data_port = default_nmea_data_port;

#define TS_ID "UNKNOWN"
#define TS_PASSWD "UNKNOWN"

volatile sig_atomic_t exit_program = 0;

int init_serial (struct termios * oldtio)
{
	/* 
	 * Open up the serial port connected to the GPS receiver
	 * save the current settins and set Baud rate to 4800 
	 * using 8N1 (8 bits, No parity, 1 stop bit)
	 *
	 * Returns the file descriptor of the open port on success
	 *
	 * Exits program if no device found 
	 *
	 */

	struct termios newtio;
	int fd;
	
	/* Open the GPS serail port */
	fd = open(nmea_data_port, O_RDONLY | O_NOCTTY | O_NDELAY);
	
	/* Exit on failure */
	if (fd < 0) 
	{
		perror(nmea_data_port); 
		exit(-1); 
	}
	else
	{
		fcntl(fd, F_SETFL, 0); 	// Prevents waiting for chars on READ
	}

	/* Save the current port settings */
	tcgetattr (fd, oldtio);	

	/* Get current port settings */
	tcgetattr (fd, &newtio);	

	/* Set output baud rate */
	cfsetispeed(&newtio, BAUDRATE);
	cfsetospeed(&newtio, BAUDRATE);

	/* Enable the reciever and set local */
	newtio.c_cflag |= (CLOCAL | CREAD);

	/* Set 8N1 */
	newtio.c_cflag &= ~PARENB;
	newtio.c_cflag &= ~CSTOPB;
	newtio.c_cflag &= ~CSIZE;
	newtio.c_cflag |= CS8;

	/* Set raw input */
	newtio.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);

	/* Ignore CR NMEA does CR LF at end of each sentence */
	newtio.c_iflag |= (IGNCR);
        
	/* Set raw output (mute point as we don't send stuff to gps in here) */
	newtio.c_oflag &= ~OPOST;

	/* Save the new settings to the port */
	tcsetattr(fd, TCSANOW, &newtio);  // Set the changes immediately

	return (fd);
}

void release_serial (int fd, struct termios * oldtio)
{

	/* Set the port back to its original settings */
	tcsetattr (fd, TCSANOW, oldtio); 

	/* Close the serial port */
	close(fd);
}

void read_string (int fd, char * buffer)
{

	/*
	 * Read a whole string from the GPS device into the buffer
	 * making sure that the size of the buffer is not exceeded
	 * in the process.
	 */

	int eos = 0;           // End of string flag
	int buffer_pos = 0;    // Buffer position
	char c;		       // Char read from port

	while (!eos && (buffer_pos < MAX_NMEA_STRING) )
	{

		if (read(fd, &c,1)) // Read in a char; returns true if char waiting
		{
			*buffer = c;     // Assign char to NMEA string buffer;	

			buffer ++;       // Increment the buffer 
			buffer_pos ++;   // Increment the buffer 

			if (c == '\n')   // Check for end of string
			   eos = 1; 

		} 

	}

	*buffer = '\0';    // Terminate the end of the buffer 

}

int check_string (const char * string)
{

	/*
	 * Check the string contains only valid NMEA chars
	 * and count the number of fields.
	 *
	 * Returns number of fields on success 0 on failure
	 */

	int char_count = 0;
	int field_count = 0;
	int invalid_chars = FALSE;
	char c = 32;

	while ((char_count < MAX_NMEA_STRING) && (c != '\0'))
	{
		c = string[char_count];
	
		if (c != '\0') 
		{
			/* Check if chars are valid */
			if ( (c < 0x20 || c > 0x7e) && (c !=0xa ) ) 
				invalid_chars = TRUE;
			
			/* Increment the field count if comma */
			if (c == ',') field_count ++;
		}

		char_count ++;
	}

	if (invalid_chars) 
		return 0; 
	else
		return field_count ++;

}

void parse_GGA (struct track_string * ts, const char * string)
{
	/*
	 * Efficiency single pass of string
	 */

	int char_count = 0;  // Char counter for parsing string
	int field_count = 1; // We start in the first field
	int field_pos = 0;   // Char counter within field
	char c = ' ';        // Must be init to anything other than \0
	
	while ((char_count < MAX_NMEA_STRING) && (c != '\0'))
	{
		c = string[char_count];

		if (c != '\0') 
		{
			switch (field_count)
			{
				case 1: // Sentence type ($GPGGA in this case)
				break;

				case 2: // Time ignore any values after dp
					if (c == ',')
					{
						ts->time[field_pos] = '\0';
					}
					else
					{
						if (c == '.')
						{
							ts->time[field_pos] = '\0';
						}
						else
						{
							ts->time[field_pos] = c;
						}
						field_pos ++;
					}
				break;

				case 3: // Latitude
					if (c == ',')
					{
						ts->lat[field_pos] = '\0';
					}
					else
					{
						ts->lat[field_pos] = c;
						field_pos ++;
					}
				break;

				case 4: // Latitude Direction
					if (c == 'N' || c == 'S')
					{
						if (ts_pos_format == ts_hms_fmt)
						{
							strncat (ts->lat,&c,1);
						}
					}
				break;

				case 5: // Longitude
					if (c == ',')
					{
						ts->lng[field_pos] = '\0';
					}
					else
					{
						ts->lng[field_pos] = c;
						field_pos ++;
					}
				break;

				case 6: // Longitude Direction
					if (c == 'E' || c == 'W')
					{
						if (ts_pos_format == ts_hms_fmt)
						{
							strncat (ts->lng,&c,1);
						}
					}
				break;

				case 7: // Fix quality
				break;

				case 8: // Satellites being tracked
				break;

				case 9: // Hoz. dilution of position
				break;

				case 10: // Altitude in meters with respect to sea level
					if (c == ',')
					{
						ts->alt[field_pos] = '\0';
					}
					else
					{
						ts->alt[field_pos] = c;
						field_pos ++;
					}
				break;

				case 11: // Altitude units (Meters)
				break;

				case 12: // Height of geiod
				break;

				case 13: // Height Units (Meters)
				break;

				case 14: // Time in seconds since last DGPS update (empty field)
				break;

				case 15: // DGPS station ID number (empty field)
				break;

				case 16: // Checksum
				break;

			}

			/* Increment the field count if comma 
			 * and set the field position to 0 */
			if (c == ',') 
			{
				field_count ++;
				field_pos = 0;
			}
		}
		char_count ++;
	}

	/* 
	 * Set the status for the track string
	 * NB we need to parse a GGA and a RMC
	 * NMEA string before the track string 
	 * is valid.
	 */

	if (ts->status == ts_printed)
	{
		ts->status = ts_pending;
	}
	else
	{
		ts->status = ts_ready;
	}

}

void parse_RMC (struct track_string * ts, const char * string)
{
	/*
	 * Efficiency single pass of string
	 */

	int char_count = 0;  // Char counter for parsing string
	int field_count = 1; // We start in the first field
	int field_pos = 0;   // Char counter within field
	float speed = 0;     // Used to convert speed in to km/h
	char c = ' ';        // Must be init to anything other than \0

	while ((char_count < MAX_NMEA_STRING) && (c != '\0'))
	{
		c = string[char_count];
	
		if (c != '\0') 
		{
			switch (field_count)
			{
				case 1: // Sentence type ($GPRMC in this case)
				break;

				case 2: // Time ignore any values after dp
					if (c == ',')
					{
						ts->time[field_pos] = '\0';
					}
					else
					{
						if (c == '.')
						{
							ts->time[field_pos] = '\0';
						}
						else
						{
							ts->time[field_pos] = c;
						}
						field_pos ++;
					}
				break;

				case 4: // Latitude
					if (c == ',')
					{
						ts->lat[field_pos] = '\0';
					}
					else
					{
						ts->lat[field_pos] = c;
						field_pos ++;
					}
				break;

				case 5: // Latitude Direction
					if (c == 'N' || c == 'S')
					{
						if (ts_pos_format == ts_hms_fmt)
						{
							strncat (ts->lat,&c,1);
						}
					}
				break;

				case 6: // Longitude
					if (c == ',')
					{
						ts->lng[field_pos] = '\0';
					}
					else
					{
						ts->lng[field_pos] = c;
						field_pos ++;
					}
				break;

				case 7: // Longitude Direction
					if (c == 'E' || c == 'W')
					{
						if (ts_pos_format == ts_hms_fmt)
						{
							strncat (ts->lng,&c,1);
						}
					}
				break;

				case 8: // Speed
					if (c == ',')
					{
						ts->spd[field_pos] = '\0';

						/* Convert Speed from knots to km/s */
						speed = atof(ts->spd);
						speed = speed * 1.852;
						sprintf( ts->spd, "%0.2fkm/h", speed );

					}
					else
					{
						ts->spd[field_pos] = c;
						field_pos ++;
					}
				break;

				case 9: // Heading
					if (c == ',')
					{
						ts->hdg[field_pos] = '\0';
					}
					else
					{
						ts->hdg[field_pos] = c;
						field_pos ++;
					}
				break;

				case 10: // Date
					if (c == ',')
					{
						ts->date[field_pos] = '\0';
					}
					else
					{
						ts->date[field_pos] = c;
						field_pos ++;
					}
				break;

				case 11: // Magnetic Variation
				break;

				case 12: // Checksum
				break;

			}

			/* Increment the field count if comma 
			 * and set the field position to 0 */
			if (c == ',') 
			{
				field_count ++;
				field_pos = 0;
			}
		}
		char_count ++;
	}

	/* 
	 * Set the status for the track string
	 * NB we need to parse a GGA and a RMC
	 * NMEA string before the track string 
	 * is valid.
	 */

	if (ts->status == ts_printed)
	{
		ts->status = ts_pending;
	}
	else
	{
		ts->status = ts_ready;
	}

}

void parse_string (struct track_string * ts, char * string)
{

	/*
	 * Check for GGA and RMC sentences and use data
	 * to fill the track_string structure
	 *
	 */

	int fields = 0;

	if ( (fields = check_string(string)) )
	{
		if ( strncmp("$GPGGA", string,6) == 0 )
		{
			parse_GGA (ts, string);
		}

		if ( strncmp("$GPRMC", string,6) == 0 )
		{
			parse_RMC (ts, string);
		}
	}
	else
	{
		printf ("Invalid string\n");
	}

}

void catch_signal (int sig)
{

	/* 
	 * Capture a signal and exit cleanly
	 * by setting an exit program flag
	 *
	 * Done to prevent ^c halting program
	 * and not restoring the initial port
	 * settings.
	 */

	/* Set the exit program flag */
	exit_program = TRUE;

	/* Capture any further signals */
	signal (sig, catch_signal);

}

void init_ts (struct track_string * ts)
{
	strcpy (ts->id,TS_ID);
	ts->time[0]='\0';
	ts->date[0]='\0';
	ts->lat[0]='\0';
	ts->lng[0]='\0';
	ts->alt[0]='\0';
	ts->hdg[0]='\0';
	ts->spd[0]='\0';
	strcpy (ts->pass,TS_PASSWD);
	ts->status=ts_printed;
}

void print_ts (struct track_string * ts)
{

	if (  ts_output_format == ts_url_fmt )
	{
		printf ("vehicle=%s&time=%s&lat=%s&lon=%s&alt=%s&heading=%s&speed=%s&pass=%s\n",
			ts->id, ts->time, ts->lat, ts->lng, ts->alt, ts->hdg, ts->spd, ts->pass 
			);
	}
	else
	{
		printf ("N:%s,T:%s,Lat:%s,Lon:%s,Alt:%s,H:%s,S:%s,P:%s\n",
	 		ts->id, ts->time, ts->lat, ts->lng, ts->alt, ts->hdg, ts->spd, ts->pass 
	 		);
	}

	/* Set the track string status to printed */
	ts->status = ts_printed;

}

void print_raw_ts (struct track_string * ts)
{
		printf ("id:%s\n", ts->id);
		printf ("time:%s\n", ts->time);
		printf ("date:%s\n", ts->date);
		printf ("latitude:%s\n", ts->lat);
		printf ("longitude:%s\n", ts->lng);
		printf ("altitude:%s\n", ts->alt);
		printf ("heading:%s\n", ts->hdg);
		printf ("speed:%s\n", ts->spd);
		printf ("password:%s\n", ts->pass);
		printf ("status:%d\n", ts->status);
}

int main (int argc, char *argv[])
{
	int result = 0;                        // Return result
	int fd = 0;                            // File descriptor
	char buffer [MAX_NMEA_STRING] = "\n\0";  // NMEA v3.01 MAX = 82;
	struct termios oldtio;                 // Storage for inital termios settings
	struct track_string ts;

	/* Init the track_string data structure */
	init_ts (&ts);

	opterr=0;
	int c;

	while ((c = getopt (argc, argv, "def:i:p:P:r")) != -1) 
	{
		switch (c)
		{

		case 'd': // Set decimal format
			ts_pos_format = ts_dec_fmt;
		break;

		case 'e': // Set email format
			ts_output_format = ts_email_fmt;
		break;

		case 'f': // Set frequency
			ts_frequency = atoi(optarg);
		break;

		case 'i': // Set the unit ID
			strcpy(ts.id,optarg);
		break;

		case 'p': // GPS data port
			nmea_data_port = optarg;
		break;

		case 'P': // mysql password 
			strcpy(ts.pass,optarg);
		break;

		case 'r': // Keep producing track strings
			ts_repeat = TRUE;
		break;

		case '?':
			if (optopt == 'c')
				fprintf (stderr, 
					"Option -%c requires an argument.\n", 
					optopt);
			else if (isprint (optopt))
				fprintf (stderr, "Unknown option `-%c'.\n", 
					optopt);
			else
				fprintf (stderr, 
					"Unknown option character `\\x%x'.\n",
					optopt);
			return 1;
		default:
			abort ();
		}
	}


	/* Init the serial port to 4800 8N1 */
 	fd = init_serial (&oldtio);


	/* Capture termiation signals */
	if (signal (SIGINT, catch_signal) == SIG_IGN)
      		signal (SIGINT, SIG_IGN);
        if (signal (SIGHUP, catch_signal) == SIG_IGN)
		signal (SIGHUP, SIG_IGN);
	if (signal (SIGTERM, catch_signal) == SIG_IGN)
		signal (SIGTERM, SIG_IGN);


	/* Read and discard the first line as it is likely to be partial data */
	read_string (fd, buffer);       

	/* Read NMEA stings until exit program flag set */
	while ( !exit_program )
	{
		read_string (fd, buffer);
		parse_string (&ts, buffer);

		// DEBUG printf ("buffer:%s", buffer);
		// DEBUG print_raw_ts (&ts);
		if (ts.status == ts_ready) 
		{
			/* Print the track string only if lat & lng data available */
			if ((strlen(ts.lat) > 0) && (strlen(ts.lng) > 0)) 
			{
				print_ts (&ts);
			}
			else
			{
				result = 2; // Exit with an error no lat & lng
			}

			if ( ! ts_repeat) exit_program = TRUE;
		}

	}

	/* Set port setting back to original settings and close port */
 	release_serial (fd, &oldtio);

 	return (result);

}
